# -*- coding: utf-8 -*-
"""tokenizer."""


def cleaning(text):
    """
    Clean the text from special characters.

    :param text: input text
    :type text: str
    :return: cleaned text
    """
    return text.replace('*', '')\
        .replace('(w)', '')\
        .replace('[', '')\
        .replace(']', '')\
        .replace('<', '')\
        .replace('>', '')


def preprocess(sentence):
    """
    Preprocess the sentence.

    :param sentence: input sentence
    :type sentence: str
    :return: preprocessed sentence
    """
    sentence = sentence.replace('-', ' ')
    return cleaning(sentence)


def run(sentence):
    """
    Tokenize the sentence.

    :param sentence: input sentence
    :type sentence: str
    :return: list of tokenized words
    """
    sentence = preprocess(sentence)
    result = []
    for index, word in enumerate(sentence.split()):
        data = {'id': index,
                'text': word}
        result.append(data)
    return result



# -*- coding: utf-8 -*-
"""pipeline."""
from .word_stemmer import run as word_stemmer_run
from .p2t import run as p2t_run
from .tokenizer import run as tokenizer_run
from .pos_tagger import run as pos_tagger_run
from .params import Task, INVALID_TASKS

TASK2FUNCTION = {
    Task.TOKENIZER.value: tokenizer_run,
    Task.P2T.value: p2t_run,
    Task.LEMMA.value: word_stemmer_run,
    Task.POS.value: pos_tagger_run
}


def pipeline(tasks, sentence):
    """
    Pipeline function to run multiple tasks on a sentence.

    :param tasks: List of tasks to run on the sentence
    :type tasks: list of Tasks
    :param sentence: Input sentence
    :type sentence: str
    :return: Dictionary containing the output of each task
    """
    unsupported_tasks = [x for x in tasks if x not in Task]
    if unsupported_tasks:
        raise ValueError(INVALID_TASKS.format(unsupported_tasks=', '.join(unsupported_tasks)))

    result = {}
    for task in tasks:
        task_output = TASK2FUNCTION[task.value](sentence)
        result[task.value] = task_output
    return result



# -*- coding: utf-8 -*-
"""transliteration."""
from .word_stemmer import find_root
from .data import PREFIXES, POSTFIXES
from .data import TRANSLITERATION_TO_TRANSCRIPTION_RULES
from .data import STEMS


def oov_transliterate(word):
    """
    Transliterates a word that is not in the STEMS dictionary.

    :param word: word to be transliterated
    :type word: str
    :return: str
    """
    first_letter = True
    new_word = ''
    for char in word:
        if first_letter:
            new_word += TRANSLITERATION_TO_TRANSCRIPTION_RULES[char][0]
        else:
            new_word += TRANSLITERATION_TO_TRANSCRIPTION_RULES[char][-1]
    return new_word


def get_transliteration(stem, post_fixes_list, pre_fixes_list):
    """
    Transliteration of a word given its stem and affixes.

    :param stem: word stem
    :type stem: str
    :param post_fixes_list: list of post-fixes
    :type post_fixes_list: list
    :param pre_fixes_list: list of pre-fixes
    :type pre_fixes_list: list
    :return: str
    """
    if post_fixes_list:
        post_fixes = ''.join(POSTFIXES[post_fix] for post_fix in reversed(post_fixes_list))
    else:
        post_fixes = ''

    if pre_fixes_list:
        pre_fixes = ''.join(PREFIXES[pre_fix] for pre_fix in reversed(pre_fixes_list))
    else:
        pre_fixes = ''

    for stem_, transit_ in zip(STEMS['stem'], STEMS['stem_translit']):
        if stem_ == stem:
            return str(pre_fixes) + str(transit_) + str(post_fixes)
    # Return None if no match is found to ensure consistent function behavior
    return oov_transliterate(stem)


def run(sentence):
    """
    Run the transliteration on a sentence.

    :param sentence: sentence to be processed
    :type sentence: str
    :return: list
    """
    result = []
    for word in sentence.split():
        roots = find_root(word)
        data = {'text': word,
                'transliteration': get_transliteration(**roots)}
        result.append(data)
    return result



# -*- coding: utf-8 -*-
"""data."""
import os
import pandas as pd


def get_data_path(data_name):
    """
    Return data file path.

    :param data_name: data file name
    :type data_name: str
    :return: direct path to data file
    """
    cd, _ = os.path.split(__file__)
    return os.path.join(cd, "files", data_name)


ROOTS = list(pd.read_json(get_data_path('roots.json'))['data'])

STEMS = pd.read_csv(get_data_path('stems.csv'))

EMISSION = pd.read_csv(get_data_path("emission.csv"), index_col=0)

TRANSITION = pd.read_csv(get_data_path("transition.csv"), index_col=0)


PREFIXES = {
    'a': '',
    'an': '',
    'abē': 'ʾp̄y',
    'duš': 'dwš',
    'ham': 'hm',
    'hām': 'hʾm',
    'hu': 'hw',
    'jud': "ywdt'",
}

POSTFIXES = {
    'am': 'm',
    'ēd': "yt'",
    'ēnd': 'd',
    'ag': "k'",
    'āg': "'k",
    'agān': "kʾn",
    'an': 'n',
    'ān': "ʾn",
    'ānag': "ʾnk",
    'ār': "ʾl",
    'āwand': "ʾwnd",
    'bān': "bʾn",
    'bad': 'pt',
    'bed': 'pt',
    'dān': "dʾn",
    'ēn': "yn",
    'endag': "yndk",
    'estān': "stʾn",
    'gāh': "gʾs",
    'gānag': "kʾnk",
    'gar': "kl",
    'gār': "kʾl",
    'gēn': "kyn",
    'īg': "yk",
    'īgān': "ykʾn",
    'īh': "yh",
    'īha': "yhʾ",
    'īhā': "yh'",
    'išn': "šn'",
    'išt': "ʾšt'",
    'īzag': "yck",
    'om': "wm",
    'mand': "mnd",
    'ōmand': "'wmnd",
    'rōn': "lwn",
    'tar': "tl",
    'dar': "tl",
    'dār': "tʾl",
    'tom': "twm",
    'dom': "twm",
    'war': "wl",
    'wār': "wʾl",
    'zār': "cʾl",
    'īd': "yt",
}


TRANSLITERATION_TO_TRANSCRIPTION_RULES = {
    "t": ("t"),
    "š": ("š"),
    "č": ("c"),
    "p": ("p"),
    "f": ("f"),
    "s": ("s"),
    "h": ("s"),
    "l": ("l"),
    "k": ("k"),
    "z": ("z"),
    "r": ("r"),
    "n": ("n"),
    "w": ("w", "wb"),
    "m": ("m", "nb"),
    "y": ("y"),
    "g": ("g", "k"),
    "d": ("d"),
    "b": ("b", "p"),
    "a": ("h"),
    "ā": ("h"),
    "x": ("h"),
    "j": ("y"),
    "ē": ("i"),
    "e": ("i"),
    "i": ("i"),
    "c": ("c"),
    "u": ('w'),
    "ī": ('y'),
    "ō": ('w'),
    "ū": ('w'),
    "ǰ": ('c'),
}



# -*- coding: utf-8 -*-
"""parsipy module."""
from .params import Task, PARSIPY_VERSION, POSTaggerModel
from .word_stemmer import find_root
from .tokenizer import preprocess, cleaning
from .pos_tagger import POSTagger
from .p2t import oov_transliterate, get_transliteration
from .pipeline import pipeline

__version__ = PARSIPY_VERSION



# -*- coding: utf-8 -*-
"""parameters and constants."""
from enum import Enum

PARSIPY_VERSION = "0.1"

INVALID_TASKS = "Sorry, the following tasks are not supported yet: {unsupported_tasks}"

POS_MAPPING = {
    10: 'N',     # Noun
    20: 'ADJ',   # ADJective
    21: 'DET',   # DETerminer
    30: 'V',     # Verb
    40: 'ADV',   # ADVerb
    50: 'PRON',  # PRONoun
    60: 'PREP',  # PREPosition
    61: 'POST',  # POSTposition
    63: 'CONJ',  # CONJunction
    64: 'EZ',    # EZafeh
    70: 'NUM',   # NUMber
    80: 'PART'   # PARTicle
}


class POSTaggerModel(Enum):
    """POSTagger models."""

    RULE_BASED = "rule_based"


class Task(Enum):
    """Tasks supported by Parsipy."""

    LEMMA = "lemma"
    POS = "POS"
    P2T = "P2T"
    TOKENIZER = "tokenizer"



# -*- coding: utf-8 -*-
"""POS tagger."""
from .data import EMISSION, TRANSITION
from .params import POSTaggerModel, POS_MAPPING


class POSTaggerRuleBased:
    """Rule-based POS tagger using Viterbi algorithm."""

    def __init__(self, smoothing_const=0.0001):
        """
        Initialize the POS tagger.

        :param smoothing_const: Smoothing constant for unknown words
        :type smoothing_const: float
        """
        self.smoothing_const = smoothing_const
        self.trans_df = TRANSITION
        self.emission_df = EMISSION
        self.tagset = sorted(set(POS_MAPPING.values()))

    def viterbi(self, words, smoothing=True):
        """
        Viterbi decoding algorithm for POS tagging.

        :param words: List of words
        :type words: list
        :param smoothing: Whether to use smoothing for unknown words
        :type smoothing: bool
        :return: List of tuples containing word and its POS tag
        """
        viterbi_list = []
        for key, word in enumerate(words):
            if word.endswith('īhā'):
                viterbi_list.append('ADV')
                continue
            max_prob = -1
            best_tag = None
            for tag in self.tagset:
                if key == 0:
                    transition_p = self.trans_df.loc['<S>', tag]
                else:
                    last_state = viterbi_list[-1]
                    transition_p = self.trans_df.loc[last_state, tag]

                if word in self.emission_df.index:
                    emission_p = self.emission_df.loc[word, tag]
                elif smoothing:
                    emission_p = self.smoothing_const
                else:
                    emission_p = 0
                prob = emission_p * transition_p
                if prob >= max_prob:
                    max_prob = prob
                    best_tag = tag
            viterbi_list.append(best_tag)
        word_tag = list(zip(words, viterbi_list))
        return word_tag


class POSTagger:
    """POS Tagger class."""

    def __init__(self, model=POSTaggerModel.RULE_BASED):
        """
        Initialize the POS tagger.

        :param model: POS tagger model
        :type model: POSTaggerModel
        """
        self.model = model
        if model == POSTaggerModel.RULE_BASED:
            self.tagger = POSTaggerRuleBased()
        else:
            pass

    def tag(self, sentence):
        """
        Return part-of-speech information of the words in the sentence.

        :param sentence: Input sentence
        :type sentence: str
        :return: List of dictionaries containing word and its POS tag
        """
        if self.model == POSTaggerModel.RULE_BASED:
            result = []
            for word, tag in self.tagger.viterbi(("<S> " + sentence).split(), smoothing=True):
                result.append({'text': word, 'POS': tag})
            return result[1:]
        else:
            pass


def run(sentence):
    """
    Run the POS tagger on the input sentence.

    :param sentence: Input sentence
    :type sentence: str
    :return: List of dictionaries containing word and its POS tag
    """
    tagger = POSTagger()
    return tagger.tag(sentence=sentence)



# -*- coding: utf-8 -*-
"""word stemmer."""
import copy
from .data import PREFIXES, POSTFIXES
from .data import ROOTS


def find_root(word):
    """
    Find the root of a word.

    :param word: Input word
    :type word: str
    :return: Dictionary containing the root of the word and the list of prefixes and postfixes
    """
    candidates = [{"stem": word, "post_fixes_list": [], "pre_fixes_list": []}]

    visited = list()
    while candidates:
        candidate = candidates.pop(0)
        if candidate['stem'] in ROOTS:
            return candidate

        if candidate not in visited:
            visited.append(candidate)

        for postfix in POSTFIXES.keys():
            if candidate['stem'].endswith(postfix):
                postfixes_list = candidate['post_fixes_list'].copy()
                prefix_list = candidate['pre_fixes_list'].copy()
                postfixes_list.append(postfix)
                stem = copy.deepcopy(candidate['stem'])[:-len(postfix)]
                new_candidate = {'stem': stem,
                                 'post_fixes_list': postfixes_list, 'pre_fixes_list': prefix_list}
                if new_candidate not in visited:
                    candidates.append(new_candidate)

        for prefix in PREFIXES.keys():
            if candidate['stem'].startswith(prefix):
                postfixes_list = candidate['post_fixes_list'].copy()
                prefix_list = candidate['pre_fixes_list'].copy()
                prefix_list.append(prefix)
                stem = copy.deepcopy(candidate['stem'])[len(prefix):]
                new_candidate = {'stem': stem,
                                 'post_fixes_list': postfixes_list, 'pre_fixes_list': prefix_list}
                if new_candidate not in visited:
                    candidates.append(new_candidate)
    return {'stem': word, 'post_fixes_list': [], 'pre_fixes_list': []}


def run(sentence):
    """
    Run the word stemmer on a sentence.

    :param sentence: Input sentence
    :type sentence: str
    :return: List of dictionaries containing the root of each word in the sentence
    """
    result = []
    for word in sentence.split():
        roots = find_root(word)
        data = {'text': word,
                'stem': roots['stem']}
        result.append(data)
    return result
